@@ -2,7 +2,7 @@ package Importer;
 use strict;
 use warnings;
 
-our $VERSION = 0.005;
+our $VERSION = 0.006;
 
 my %SIG_TO_SLOT = (
     '&' => 'CODE',
@@ -12,43 +12,10 @@ my %SIG_TO_SLOT = (
     '*' => 'GLOB',
 );
 
-my %IMPORTED;
+our %IMPORTED;
 
 # This will be used to check if an import arg is a version number
-my %NUMERIC = map { $_ => 1 } 0 .. 9;
-
-# If a consumer just wants subs then we can optimize the import. This is used
-# as a lookup table to find non-optimal sigils. Can't just look for '&' since a
-# sub can be listed without a sigil, so alpha-numerics may also be checked
-# against thi stable, and we want those to be considered optimal.
-my %NON_OPTIMAL = ( '$' => 1, '@' => 1, '%' => 1, '*' => 1 );
-
-###########################################################################
-#
-# This is a method intended to be used as 'import' by packages that want to
-# export on use.
-#
-
-our @EXPORT_OK = qw/exporter_import/;
-sub exporter_import {
-    my $from = shift;
-
-    my @caller = caller(0);
-
-    return unless @_;
-
-    my $file = _mod_to_file($from);
-    _load_file(\@caller, $file) unless $INC{$file};
-
-    return if _optimal_import($from, $caller[0], \@caller, @_);
-
-    my $self = __PACKAGE__->new(
-        from   => $from,
-        caller => \@caller,
-    );
-
-    $self->do_import($caller[0], @_);
-}
+my %NUMERIC = map +($_ => 1), 0 .. 9;
 
 ###########################################################################
 #
@@ -75,7 +42,16 @@ sub import {
     my $file = _mod_to_file($from);
     _load_file(\@caller, $file) unless $INC{$file};
 
-    return if _optimal_import($from, $caller[0], \@caller, @args);
+    no strict 'refs';
+    no warnings 'once';
+    _optimal_import($from, $caller[0], \@caller, @args)
+        and return
+        unless defined *{"$from\::IMPORTER_MENU"}{CODE} # Origin package has a custom menu
+            || defined *{"$from\::EXPORT_FAIL"}{ARRAY}  # Origin package has a failure handler
+            || defined *{"$from\::EXPORT_GEN"}{HASH}    # Origin package has generators
+            || defined *{"$from\::EXPORT_ANON"}{HASH};  # Origin package has anonymous exports
+    use strict 'refs';
+    use warnings 'once';
 
     my $self = $class->new(
         from   => $from,
@@ -115,7 +91,16 @@ sub import_into {
     my $file = _mod_to_file($from);
     _load_file(\@caller, $file) unless $INC{$file};
 
-    return if _optimal_import($from, $into, \@caller, @args);
+    no strict 'refs';
+    no warnings 'once';
+    _optimal_import($from, $into, \@caller, @args)
+        and return
+        unless defined *{"$from\::IMPORTER_MENU"}{CODE} # Origin package has a custom menu
+            || defined *{"$from\::EXPORT_FAIL"}{ARRAY}  # Origin package has a failure handler
+            || defined *{"$from\::EXPORT_GEN"}{HASH}    # Origin package has generators
+            || defined *{"$from\::EXPORT_ANON"}{HASH};  # Origin package has anonymous exports
+    use strict 'refs';
+    use warnings 'once';
 
     my $self = $class->new(
         from   => $from,
@@ -249,21 +234,25 @@ sub get_caller {
     }
 
     # Fallback
-    return [caller(1)];
+    return [caller(0)];
 }
 
 sub croak {
     my $self = shift;
     my ($msg) = @_;
     my $caller = $self->get_caller;
-    die "$msg at $caller->[1] line $caller->[2].\n";
+    my $file = $caller->[1] || 'unknown file';
+    my $line = $caller->[2] || 'unknown line';
+    die "$msg at $file line $line.\n";
 }
 
 sub carp {
     my $self = shift;
     my ($msg) = @_;
     my $caller = $self->get_caller;
-    warn "$msg at $caller->[1] line $caller->[2].\n";
+    my $file = $caller->[1] || 'unknown file';
+    my $line = $caller->[2] || 'unknown line';
+    warn "$msg at $file line $line.\n";
 }
 
 sub menu {
@@ -288,13 +277,18 @@ sub reload_menu {
     my $from = $self->from;
 
     my ($export, $export_ok, $export_tags, $export_fail, $generate, $export_gen, $export_anon, $new_style);
-    if ($from->can('IMPORTER_MENU')) {
+
+    no strict 'refs';
+    no warnings 'once';
+    if (my $menu_sub = *{"$from\::IMPORTER_MENU"}{CODE}) {
+        use strict 'refs';
+        use warnings 'once';
         # Hook, other exporter modules can define this method to be compatible with
         # Importer.pm
 
         $new_style = 1;
 
-        my %got = $from->IMPORTER_MENU($into, $self->get_caller);
+        my %got = $from->$menu_sub($into, $self->get_caller);
         $export      = $got{export}      || [];
         $export_ok   = $got{export_ok}   || [];
         $export_tags = $got{export_tags} || {};
@@ -310,8 +304,6 @@ sub reload_menu {
         $export_gen ||= {};
     }
     else {
-        no strict 'refs';
-        no warnings 'once';
         $export      = \@{"$from\::EXPORT"};
         $export_ok   = \@{"$from\::EXPORT_OK"};
         $export_tags = \%{"$from\::EXPORT_TAGS"};
@@ -319,6 +311,8 @@ sub reload_menu {
         $export_gen  = \%{"$from\::EXPORT_GEN"};
         $export_anon = \%{"$from\::EXPORT_ANON"};
     }
+    use strict 'refs';
+    use warnings 'once';
 
     $generate ||= sub {
         my $symbol = shift;
@@ -338,7 +332,6 @@ sub reload_menu {
     for my $sym (@$export, @$export_ok, keys %$export_gen, keys %$export_anon) {
         my ($sig, $name) = ($sym =~ m/^(\W?)(.*)$/);
         $sig ||= '&';
-        my $slot = $SIG_TO_SLOT{$sig};
 
         $lookup->{"${sig}${name}"} = 1;
         $lookup->{$name} = 1 if $sig eq '&';
@@ -348,7 +341,15 @@ sub reload_menu {
 
         no strict 'refs';
         no warnings 'once';
-        $exports->{"${sig}${name}"} = $export_anon->{$sym} || ($slot eq 'SCALAR' ? \${"$from\::$name"} : *{"$from\::$name"}{$slot});
+        my $fqn = "$from\::$name";
+        $exports->{"${sig}${name}"} = $export_anon->{$sym} || (
+            $sig eq '&' ? \&{$fqn} :
+            $sig eq '$' ? \${$fqn} :
+            $sig eq '@' ? \@{$fqn} :
+            $sig eq '%' ? \%{$fqn} :
+            $sig eq '*' ? \*{$fqn} :
+            die "This should not happen"
+        );
     }
 
     my $f_import = $new_style || $from->can('import');
@@ -527,8 +528,10 @@ sub { *{"$into\\::\$_[0]"} = \$_[1] }
         $self->croak("$from does not export $symbol") unless $ref || $menu->{lookup}->{"${sig}${name}"};
         next unless $ref;
 
+        my $type = ref($ref);
+        $type = 'SCALAR' if $type eq 'REF';
         $self->croak("Symbol '$sig$name' requested, but reference (" . ref($ref) . ") does not match sigil ($sig)")
-            if $ref && ref($ref) ne $SIG_TO_SLOT{$sig};
+            if $ref && $type ne $SIG_TO_SLOT{$sig};
 
         # If they directly renamed it then we assume they want it under the new
         # name, otherwise excludes get kicked. It is useful to be able to
@@ -578,46 +581,34 @@ require \$file;
 sub _optimal_import {
     my ($from, $into, $caller, @args) = @_;
 
-    my %new_style = $from->can('IMPORTER_MENU') ? $from->IMPORTER_MENU : ();
-
-    my %final;
     no strict 'refs';
-    return 0 if $new_style{export_fail} || @{"$from\::EXPORT_FAIL"};
-    @args = @{$new_style{export} || "$from\::EXPORT"} unless @args;
-    my %allowed = map +($_ => 1), @{$new_style{export} || "$from\::EXPORT"}, @{$new_style{export_ok} || "$from\::EXPORT_OK"};
-    use strict 'refs';
 
-    for my $arg (@args) {
-        # Get sigil, or first letter of name
-        my $sig = substr($arg, 0, 1);
+    # Default to @EXPORT
+    @args = @{"$from\::EXPORT"} unless @args;
 
-        # Return if non-sub sigil
-        return 0 if $NON_OPTIMAL{$sig};
+    # Subs will be listed without sigil in %allowed, all others keep sigil
+    my %allowed = map +(substr($_, 0, 1) eq '&' ? substr($_, 1) : $_ => 1),
+        @{"$from\::EXPORT"}, @{"$from\::EXPORT_OK"};
 
-        # Strip sigil (if sub)
-        my $name = $arg;
-        substr($name, 0, 1, '') if $sig eq '&';
+    # First check if it is allowed, stripping '&' if necessary, which will also
+    # let scalars in, we will deal with those shortly.
+    # If not allowed return 0 (need to do a heavy import)
+    # if it is allowed then see if it has a CODE slot, if so use it, otherwise
+    # we have a symbol that needs heavy due to non-sub, autoload, etc.
+    # This will not allow $foo to import foo() since '$from' still contains the
+    # sigil making it an invalid symbol name in our globref below.
+    my %final = map +(
+        ($allowed{$_} || (substr($_, 0, 1, "") eq '&' && $allowed{$_}))
+            ? ($_ => *{"$from\::$_"}{CODE} || return 0)
+            : return 0
+    ), @args;
 
-        # Check if the name is allowed (with or without sigil)
-        return 0 unless $allowed{$name} || $allowed{$arg};
-
-        no strict 'refs';
-        $final{$name} = \&{"$from\::$name"};
-    }
-
-    # This is necessary for the eval.
-    my $IMPORTED = \%IMPORTED;
-
-    # This effects the eval, which is what we want. Putting this here means it
-    # runs once, at build time. Putting it inside the eval means it runs each
-    # time _optimal_import is called, which is costly.
-    no strict 'refs';
     eval <<"    EOT" || die $@;
 # If the caller has redefine warnings enabled then we want to warn them if
 # their import redefines things.
 BEGIN { \${^WARNING_BITS} = \$caller->[9] if defined \$caller->[9] };
 #line $caller->[2] "$caller->[1]"
-(*{"$into\\::\$_"} = \$final{\$_}, push \@{\$IMPORTED->{\$into}} => \$_) for keys %final;
+(*{"$into\\::\$_"} = \$final{\$_}, push \@{\$Importer::IMPORTED{\$into}} => \$_) for keys %final;
 1;
     EOT
 }
@@ -639,14 +630,13 @@ Importer - Alternative but compatible interface to modules that export symbols.
 This module acts as a layer between L<Exporter> and modules which consume
 exports. It is feature-compatible with L<Exporter>, plus some much needed
 extras. You can use this to import symbols from any exporter that follows
-L<Exporters> specification. The exporter modules themselves do not need to use
+L<Exporter>s specification. The exporter modules themselves do not need to use
 or inherit from the L<Exporter> module, they just need to set C<@EXPORT> and/or
 other variables.
 
 =head1 *** EXPERIMENTAL ***
 
-This module is still experimental. Anything can change at any time. Testing is
-currently VERY insufficient.
+This module is still experimental. Anything can change at any time.
 
 =head1 SYNOPSYS
 
@@ -828,6 +818,8 @@ List of symbols to export. Sigil is optional for subs. Symbols listed here are
 exported by default. If possible you should put symbols in C<@EXPORT_OK>
 instead.
 
+    our @EXPORT = qw/foo bar &baz $BAT/;
+
 =head2 @EXPORT_OK
 
 This is used exactly the way L<Exporter> uses it.
@@ -835,6 +827,8 @@ This is used exactly the way L<Exporter> uses it.
 List of symbols that can be imported. Sigil is optional for subs. Symbols
 listed here are not exported by default. This is preferred over C<@EXPORT>.
 
+    our @EXPORT_OK = qw/foo bar &baz $BAT/;
+
 =head2 %EXPORT_TAGS
 
 This module supports tags exactly the way L<Exporter> does.
@@ -849,9 +843,11 @@ This is used exactly the way L<Exporter> uses it.
 
 Use this to list subs that are not available on all platforms. If someone tries
 to import one of these, Importer will hit your C<< $from->export_fail(@items) >>
-callback to try to resolve the issue. See L<Exporter.pm> for documentation of
+callback to try to resolve the issue. See L<Exporter> for documentation of
 this feature.
 
+    our @EXPORT_FAIL = qw/maybe_bad/;
+
 =head2 %EXPORT_ANON
 
 This is new to this module, L<Exporter> does not support it.
@@ -925,19 +921,13 @@ name, or a caller level.
 
 =back
 
-=head1 GIVING YOUR PACKAGE AN OLD-SCHOOL IMPORT METHOD
-
-    package My::Exporter;
-    use Importer Importer => ('exporter_import' => {-as => 'import'});
-
-    ...
-
 =head1 USING WITH OTHER EXPORTER IMPLEMENTATIONS
 
 If you want your module to work with Importer, but you use something other than
 L<Exporter> to define your exports, you can make it work be defining the
 C<IMPORTER_MENU> method in your package. As well other exporters can be updated
-to support Importer by putting this sub in your package:
+to support Importer by putting this sub in your package.
+B<IMPORTER_MENU() must be defined in your package, not a base class!>
 
     sub IMPORTER_MENU {
         my $class = shift;
@@ -966,8 +956,10 @@ to support Importer by putting this sub in your package:
         return $ref;
     }
 
-All exports must be listed in either C<@EXPORT> or C<@EXPORT_OK> to be allowed.
-C<%EXPORT_TAGS>, C<@EXPORT_FAIL>, and C<\&GENERATE> are optional.
+All exports must be listed in either C<@EXPORT> or C<@EXPORT_OK>, or be keys in
+C<%EXPORT_GEN> or C<%EXPORT_ANON> to be allowed. 'export_tags', 'export_fail',
+'export_anon', 'export_gen', and 'generate' are optional. You cannot combine
+'generate' and 'export_gen'.
 
 B<Note:> If your GENERATE sub needs the C<$class>, C<$into>, or C<$caller> then
 your C<IMPORTER_MENU()> method will need to build an anonymous sub that closes
